package com.jpattern.orm;

import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.classmapper.ClassMapperBuilder;
import com.jpattern.orm.classmapper.ClassTableMapBuilder;
import com.jpattern.orm.classmapper.IClassMapper;
import com.jpattern.orm.classmapper.TableMap;
import com.jpattern.orm.classtool.IOrmClassTool;
import com.jpattern.orm.classtool.IOrmClassToolMap;
import com.jpattern.orm.classtool.NullOrmClassToolMap;
import com.jpattern.orm.classtool.OrmClassTool;
import com.jpattern.orm.classtool.OrmClassToolMap;
import com.jpattern.orm.crud.IOrmCRUDQuery;
import com.jpattern.orm.crud.OrmCRUDQueryGenerator;
import com.jpattern.orm.dialect.Dialect;
import com.jpattern.orm.exception.OrmConfigurationException;
import com.jpattern.orm.persistor.IOrmPersistor;
import com.jpattern.orm.persistor.anew.type.TypeWrapper;
import com.jpattern.orm.persistor.reflection.ReflectionPersistorGenerator;
import com.jpattern.orm.persistor.reflection.resultset.ResultSetMethodHelper;
import com.jpattern.orm.session.NullSessionProvider;
import com.jpattern.orm.session.OrmSession;
import com.jpattern.orm.session.Session;
import com.jpattern.orm.session.SessionProvider;
import com.jpattern.orm.validator.NullValidator;
import com.jpattern.orm.validator.Validator;

/**
 * 
 * @author Francesco Cina'
 *
 * 26/ago/2011
 */
public class JPOrm implements JPO {

	private static Integer JPORM_INSTANCES_COUNT = Integer.valueOf(0);
	private SessionProvider sessionProvider;
	private IOrmClassToolMap ormClassToolMap = new OrmClassToolMap(this);
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Dialect dialect;
	private final Integer instanceCount;
	private final ResultSetMethodHelper resultSetMethodHelper = new ResultSetMethodHelper();
	private Validator validator = new NullValidator();

	/**
	 * Create a new instance of JPOrm.
	 * 
	 * @param sessionProvider
	 */
	public JPOrm(final SessionProvider sessionProvider) {
		synchronized (JPORM_INSTANCES_COUNT) {
			this.instanceCount = JPORM_INSTANCES_COUNT++;
		}
		this.logger.info("Building new instance of JPO (instance [{}])", this.instanceCount);
		this.sessionProvider = sessionProvider;
		this.dialect = sessionProvider.getDBType().getDialect();
	}

	@Override
	public final Session session() {
		return new OrmSession(this.ormClassToolMap, this.sessionProvider.getSessionStrategy(), this.resultSetMethodHelper, this.validator);
	}

	@Override
	public synchronized <T> void register(final Class<T> clazz) throws OrmConfigurationException {
		try {
			if (!this.ormClassToolMap.containsTool(clazz)) {
				this.logger.info("register new class: " + clazz.getName());
				final TableMap<T> tableMap = new TableMap<T>();
				new ClassTableMapBuilder<T>(clazz, tableMap).updateTableMap();
				final IClassMapper<T> classMapper = new ClassMapperBuilder<T>(clazz, tableMap).generate();
				final IOrmPersistor<T> ormPersistor =  new ReflectionPersistorGenerator<T>(classMapper, this.resultSetMethodHelper).generate();
				final IOrmCRUDQuery ormCRUDQuery = new OrmCRUDQueryGenerator<T>(this.dialect, classMapper).generate();
				final IOrmClassTool<T> ormClassTool = new OrmClassTool<T>(classMapper, ormPersistor, ormCRUDQuery);
				this.ormClassToolMap.put(clazz, ormClassTool);
			}
		} catch (final Exception e) {
			throw new OrmConfigurationException(e);
		}
	}

	@Override
	public synchronized void destory() {
		this.sessionProvider = new NullSessionProvider();
		this.ormClassToolMap = new NullOrmClassToolMap();
	}

	@Override
	public synchronized void register(final List<Class<?>> classes) throws OrmConfigurationException {
		for (final Class<?> clazz : classes) {
			this.register(clazz);
		}
	}

	public SessionProvider getSessionProvider() {
		return this.sessionProvider;
	}

	public ResultSetMethodHelper getResultSetMethodHelper() {
		return this.resultSetMethodHelper;
	}

	@Override
	public synchronized void register(final TypeWrapper<?, ?> typeWrapper) throws OrmConfigurationException {
		try {
			this.resultSetMethodHelper.addWrappers(typeWrapper);
		} catch (final Exception e) {
			throw new OrmConfigurationException(e);
		}
	}

	@Override
	public synchronized void setValidator(final Validator validator) {
		if (validator!=null) {
			this.validator = validator;
		}
	}

}
