package com.jpattern.orm.classmapper;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.util.FieldDefaultNaming;

/**
 * 
 * @author Francesco Cina
 *
 * 22/mag/2011
 */
public class ClassMapperBuilder<T> extends ABuilder implements IClassMapperBuilder<T> {

	private final Class<T> clazz;
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final ITableMap<T> tableMap;

	public ClassMapperBuilder(final Class<T> clazz, final ITableMap<T> tableMap) {
		this.clazz = clazz;
		this.tableMap = tableMap;
	}

	/* (non-Javadoc)
	 * @see com.jpattern.orm.mapper.IClassMapperGenerator#generate()
	 */
	@Override
	public IClassMapper<T> generate() {
		this.logger.info("generate " + IClassMapper.class.getSimpleName() + " for Class " + this.clazz.getName());
		final String tableName = this.tableMap.getTableNameWithSchema();
		final String schemaName = this.tableMap.getSchemaName();
		this.logger.info("table name expected in relation with class " + this.clazz.getSimpleName() + ": " + tableName + " - schema: " + schemaName);
		final Map<String, IColumn> fieldColumnMapping = mapField(schemaName, tableName);
		final ClassMapper<T> classMapper = new ClassMapper<T>(this.clazz, this.tableMap, fieldColumnMapping);
		initializeColumnNames(classMapper, fieldColumnMapping);
		return classMapper ;
	}

	private Map<String, IColumn> mapField(final String schemaName, final String tableName) {
		final Map<String, IColumn> result = new LinkedHashMap<String, IColumn>();
		Column col = null;

		for ( final String fieldName : this.tableMap.getAllFieldNames() ) {
			if ( fieldName.length() > 0) {
				final String columnName = this.tableMap.getDBColumnName(fieldName);
				this.logger.info( "DB column [" + columnName + "]" + " will be associated with object field [" + fieldName + "]");
				col = new Column( columnName );
				col.setIdentifying( this.tableMap.getPrimaryKeyFieldNames().contains(fieldName) );
				setGetterSetter( fieldName, col );
				result.put( fieldName , col );
			} else {
				this.logger.warn( "Class fiels [" + fieldName + "]" + " is NOT associated with a DB colunm!");
			}
		}
		return result;
	}

	private void setGetterSetter(final String javaPropertyName, final Column col) {
		final List<Method> methods = Arrays.asList( this.clazz.getMethods() );
		final List<Field> fields = new ArrayList<Field>();
		fillInheritedFields(this.clazz, fields);
		//		final List<Field> fields = Arrays.asList( clazz.getDeclaredFields() );

		Field propertyField = null;
		Method getter = null;
		Method setter = null;
		String fieldName = "";
		String getterName = "";
		String setterName = "";

		for (final Field field : fields) {
			if (javaPropertyName.equals(field.getName())) {
				propertyField = field;
				fieldName = field.getName();
			}
		}

		for (final Method method : methods) {
			if (FieldDefaultNaming.getDefaultGetterName(javaPropertyName).equals(method.getName())) {
				getter = method;
				getterName = method.getName();
			}
			if (FieldDefaultNaming.getDefaultBooleanGetterName(javaPropertyName).equals(method.getName())) {
				getter = method;
				getterName = method.getName();
			}
			if (FieldDefaultNaming.getDefaultSetterName(javaPropertyName).equals(method.getName())) {
				setter = method;
				setterName = method.getName();
			}
		}

		this.logger.info("property [" + fieldName + "] associated with column [" + col.getName() + "]");
		this.logger.info("getter for property [" + fieldName + "]: [" + getterName + "]");
		this.logger.info("setter for property [" + fieldName + "]: [" + setterName + "]");

		col.setField(propertyField);
		col.setSetter(setter);
		col.setGetter(getter);
	}

	private void initializeColumnNames(final ClassMapper<T> generated, final Map<String, IColumn> fieldColumnMapping) {
		this.logger.info("Start column analisys");
		final List<String> allColumnJavaNamesList = new ArrayList<String>();
		final List<String> allNotGeneratedColumnJavaNamesList = new ArrayList<String>();
		final List<String> allGeneratedColumnJavaNamesList = new ArrayList<String>();
		final List<String> allGeneratedColumnDBNamesList = new ArrayList<String>();
		final List<String> primaryKeyColumnJavaNamesList = new ArrayList<String>();
		final List<String> primaryKeyAndVersionColumnJavaNamesList = new ArrayList<String>();
		final List<String> notPrimaryKeyColumnJavaList = new ArrayList<String>();
		for (final Entry<String, IColumn> entry : fieldColumnMapping.entrySet()) {
			final String javaFieldName = entry.getKey();
			allColumnJavaNamesList.add(javaFieldName);
			if( entry.getValue().isIdentifying() ) {
				primaryKeyColumnJavaNamesList.add(javaFieldName);
				primaryKeyAndVersionColumnJavaNamesList.add(javaFieldName);
				this.logger.info("Field [" + javaFieldName + "] will be used as a Primary Key field");
			} else {
				notPrimaryKeyColumnJavaList.add(javaFieldName);
				this.logger.info("Field [" + javaFieldName + "] will be used as a normal field");
			}
			if (this.tableMap.getGeneratorByJavaFieldName(javaFieldName).isAutoGenerated()) {
				allGeneratedColumnJavaNamesList.add(javaFieldName);
				allGeneratedColumnDBNamesList.add(entry.getValue().getName());
				this.logger.info("Field [" + javaFieldName + "] is an autogenerated field");
			} else {
				allNotGeneratedColumnJavaNamesList.add(javaFieldName);
			}

		}
		if (this.tableMap.isVersionable()) {
			primaryKeyAndVersionColumnJavaNamesList.add(this.tableMap.getVersionJavaFieldName());
		}

		generated.setAllColumnJavaNames( allColumnJavaNamesList.toArray(new String[0]) );
		generated.setAllNotGeneratedColumnJavaNames( allNotGeneratedColumnJavaNamesList.toArray(new String[0]) );
		generated.setAllGeneratedColumnJavaNames( allGeneratedColumnJavaNamesList.toArray(new String[0]) );
		generated.setAllGeneratedColumnDBNames( allGeneratedColumnDBNamesList.toArray(new String[0]) );
		generated.setNotPrimaryKeyColumnJavaNames( notPrimaryKeyColumnJavaList.toArray(new String[0]) );
		generated.setPrimaryKeyColumnJavaNames( primaryKeyColumnJavaNamesList.toArray(new String[0]) );
		generated.setPrimaryKeyAndVersionColumnJavaNames( primaryKeyAndVersionColumnJavaNamesList.toArray(new String[0]) );
	}
}
