package com.jpattern.orm.crud;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.jpattern.orm.classmapper.IClassMapper;
import com.jpattern.orm.classmapper.ITableMap;
import com.jpattern.orm.dialect.Dialect;

/**
 * 
 * @author Francesco Cina
 *
 * 22/mag/2011
 */
public class OrmCRUDQueryGenerator<T> implements IOrmCRUDQueryGenerator<T> {

	private final IClassMapper<T> classMapper;
	private boolean generatedKey = false;
	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Dialect dialect;

	public OrmCRUDQueryGenerator(final Dialect dialect, final IClassMapper<T> classMapper) {
		this.dialect = dialect;
		this.classMapper = classMapper;
	}

	/* (non-Javadoc)
	 * @see com.jpattern.orm.generator.IOrmCRUDQueryGenerator#generate()
	 */
	@Override
	public IOrmCRUDQuery generate() {
		final OrmCRUDQuery crud = new OrmCRUDQuery();
		crud.setDeleteQuery( generateDeleteQuery() );
		crud.setLoadQuery(  generateLoadQuery() );
		crud.setSaveQuery(  generateSaveQuery(false) );
		crud.setSaveQueryWithoutGenerators( generateSaveQuery(true) );
		crud.setUpdateQuery(  generateUpdateQuery() );
		crud.setBaseSelectClause(  generateBaseSelectClause() );
		crud.setBaseFromClause(  generateBaseFromClause() );
		crud.setBeanVersionQuery( generateBeanVersionQuery() );
		return crud;
	}

	private String generateBaseFromClause() {
		return this.classMapper.getTableMap().getTableNameWithSchema();
	}

	private String generateBaseSelectClause() {
		return columnToCommaSepareted( getColumnNames(this.classMapper , ICrudConstants.ROW_NAME_PREFIX_PLACEHOLDER, this.classMapper.getAllColumnJavaNames()), true );
	}

	private String generateUpdateQuery() {
		final StringBuilder builder = new StringBuilder("UPDATE ");
		builder.append(this.classMapper.getTableMap().getTableNameWithSchema());
		builder.append(" SET ");
		if ( this.classMapper.getNotPrimaryKeyColumnJavaNames().length>0 ) {
			builder.append( columnToSetClause( getColumnNames(this.classMapper, "", this.classMapper.getNotPrimaryKeyColumnJavaNames()) ));
		}
		if ( this.classMapper.getPrimaryKeyColumnJavaNames().length>0 ) {
			builder.append(" WHERE ");
			builder.append( columnToWhereClause( getColumnNames(this.classMapper, "", this.classMapper.getPrimaryKeyColumnJavaNames()) ));
		}
		final String query = builder.toString();
		this.logger.info("Generated UPDATE query for table [" + this.classMapper.getTableMap().getTableNameWithSchema() + "]:\n" + query );
		return query;
	}

	private String generateSaveQuery(final boolean ignoreGenerators) {
		final StringBuilder builder = new StringBuilder("INSERT INTO ");
		builder.append(this.classMapper.getTableMap().getTableNameWithSchema());
		builder.append(" (");
		builder.append( columnToCommaSepareted( getColumnNames(this.classMapper , "", this.classMapper.getAllColumnJavaNames()), ignoreGenerators ));
		builder.append(") VALUES (");
		builder.append( questionCommaSepareted( this.classMapper, getColumnNames(this.classMapper , "", this.classMapper.getAllColumnJavaNames()), ignoreGenerators ));
		builder.append(")");
		final String query = builder.toString();
		this.logger.info("Generated INSERT query (with generators? " + !ignoreGenerators + ") for table [" + this.classMapper.getTableMap().getTableNameWithSchema() + "]:\n" + query );
		return query;
	}

	private String generateLoadQuery() {
		final StringBuilder builder = new StringBuilder("SELECT * FROM ");
		builder.append(this.classMapper.getTableMap().getTableNameWithSchema());
		if ( this.classMapper.getPrimaryKeyColumnJavaNames().length>0 ) {
			builder.append(" WHERE ");
			builder.append( columnToWhereClause( getColumnNames(this.classMapper, "", this.classMapper.getPrimaryKeyColumnJavaNames()) ));
		}
		final String query = builder.toString();
		this.logger.info("Generated SELECT query for table [" + this.classMapper.getTableMap().getTableNameWithSchema() + "]:\n" + query );
		return query;
	}

	private String generateBeanVersionQuery() {
		String query = "";
		final ITableMap<T> tableMap = this.classMapper.getTableMap();
		if (tableMap.isVersionable()) {
			final StringBuilder builder = new StringBuilder("SELECT count(*) FROM ");
			builder.append(this.classMapper.getTableMap().getTableNameWithSchema());
			if ( this.classMapper.getPrimaryKeyColumnJavaNames().length>0 ) {
				builder.append(" WHERE ");
				final String[] pk = this.classMapper.getPrimaryKeyColumnJavaNames();
				final String[] pkAndVersion = new String[ pk.length + 1 ];
				for (int i=0 ; i<(pk.length); i++) {
					pkAndVersion[i] = pk[i];
				}
				pkAndVersion[pkAndVersion.length-1] = tableMap.getVersionJavaFieldName() ;

				builder.append( columnToWhereClause( getColumnNames(this.classMapper, "", pkAndVersion ) ));
			}
			builder.append( tableMap.getVersionLoadLockMode().getLockMode() );
			query = builder.toString();
		}
		this.logger.info("Generated query to load the version of a record for table [" + this.classMapper.getTableMap().getTableNameWithSchema() + "]:\n" + query );
		return query;
	}

	private String generateDeleteQuery() {
		final StringBuilder builder = new StringBuilder("DELETE FROM ");
		builder.append(this.classMapper.getTableMap().getTableNameWithSchema());
		if ( this.classMapper.getPrimaryKeyColumnJavaNames().length>0 ) {
			builder.append(" WHERE ");
			builder.append( columnToWhereClause( getColumnNames(this.classMapper, "", this.classMapper.getPrimaryKeyColumnJavaNames()) ));
		}
		final String query = builder.toString();
		this.logger.info("Generated DELETE query for table [" + this.classMapper.getTableMap().getTableNameWithSchema() + "]:\n" + query );
		return query;
	}


	private String questionCommaSepareted(final IClassMapper<T> classMapper, final List<String> columnNames, final boolean ignoreGenerators) {
		final StringBuilder builder = new StringBuilder();
		final int length = columnNames.size();

		if (length > 0) {
			for (int i=0; i<(length-1) ; i++) {
				final AColumnValueGenerator columnValueGenerator = classMapper.getTableMap().getGeneratorByColumnName(columnNames.get(i), ignoreGenerators);
				this.generatedKey = this.generatedKey || columnValueGenerator.isAutoGenerated();
				final String queryParameter = columnValueGenerator.insertQueryParameter(this.dialect, "?");
				if (queryParameter.length()>0) {
					builder.append( queryParameter );
					builder.append(", ");
				}
			}
			final AColumnValueGenerator columnValueGenerator = classMapper.getTableMap().getGeneratorByColumnName(columnNames.get(length-1), ignoreGenerators);
			this.generatedKey = this.generatedKey || columnValueGenerator.isAutoGenerated();
			builder.append( columnValueGenerator.insertQueryParameter(this.dialect, "?") );
		}
		return builder.toString();
	}

	private String columnToCommaSepareted(final List<String> columnNames, final boolean ignoreGenerators) {
		final StringBuilder builder = new StringBuilder();
		final int length = columnNames.size();

		if (length > 0) {
			for (int i=0; i<(length-1) ; i++) {
				final AColumnValueGenerator columnValueGenerator = this.classMapper.getTableMap().getGeneratorByColumnName(columnNames.get(i), ignoreGenerators);
				final String queryParameter = columnValueGenerator.insertColumn(this.dialect, columnNames.get(i));
				if (queryParameter.length()>0) {
					builder.append( queryParameter );
					builder.append(", ");
				}
			}
			final AColumnValueGenerator columnValueGenerator = this.classMapper.getTableMap().getGeneratorByColumnName(columnNames.get(length-1), ignoreGenerators);
			builder.append( columnValueGenerator.insertColumn(this.dialect, columnNames.get(length-1)) );
		}

		return builder.toString();
	}

	private String columnToWhereClause(final List<String> columnNames) {
		final StringBuilder builder = new StringBuilder();
		final int length = columnNames.size();

		if (length > 0) {
			for (int i=0; i<(length-1) ; i++) {
				builder.append(columnNames.get(i) + " = ? AND ");
			}
			builder.append(columnNames.get(length-1) + " = ? ");
		}
		return builder.toString();
	}

	private String columnToSetClause(final List<String> columnNames) {
		final StringBuilder builder = new StringBuilder();
		final int length = columnNames.size();

		if (length > 0) {
			for (int i=0; i<(length-1) ; i++) {
				builder.append(columnNames.get(i) + " = ? , ");
			}
			builder.append(columnNames.get(length-1) + " = ? ");
		}
		return builder.toString();
	}

	private List<String> getColumnNames(final IClassMapper<?> classMapper, final String rowNamePrefix, final String[] javaNames) {
		final List<String> result = new ArrayList<String>();
		for (final String javaName : javaNames) {
			result.add( rowNamePrefix + classMapper.getColumnWithJavaName(javaName).getName());
		}
		return result;
	}
}
