package com.jpattern.orm.persistor.anew;

import java.lang.reflect.InvocationTargetException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import com.jpattern.orm.persistor.anew.type.JdbcIO;
import com.jpattern.orm.persistor.anew.type.TypeWrapper;
import com.jpattern.orm.persistor.anew.version.VersionMath;
import com.jpattern.orm.persistor.reflection.field.GetFieldManipulator;
import com.jpattern.orm.persistor.reflection.field.SetFieldManipulator;

/**
 * 
 * @author ufo
 *
 * @param <P> the type of the bean to manipulate
 * @param <P> the type of the bean's property to manipulate
 * @param <DB> the type of the field in the {@link PreparedStatement} and {@link ResultSet}
 */
public class PropertyPersistor<BEAN, P, DB> {

	private final JdbcIO<DB> jdbcIO;
	private final TypeWrapper<P, DB> typeWrapper;
	private final GetFieldManipulator<P> getter;
	private final SetFieldManipulator<P> setter;
	private final VersionMath<DB> math;

	public PropertyPersistor (final TypeWrapper<P, DB> typeWrapper, final JdbcIO<DB> jdbcIO,
			final GetFieldManipulator<P> getter, final SetFieldManipulator<P> setter, final VersionMath<DB> math) {
		this.jdbcIO = jdbcIO;
		this.typeWrapper = typeWrapper;
		this.getter = getter;
		this.setter = setter;
		this.math = math;

	}

	/**
	 * Extract the value of a bean's property and set it to the {@link PreparedStatement}
	 * @param bean
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws SQLException
	 * @throws IllegalArgumentException
	 */
	public void setToPreparedStatement(final BEAN bean, final PreparedStatement ps, final int psIndex) throws IllegalArgumentException, SQLException, IllegalAccessException, InvocationTargetException {
		this.jdbcIO.setValueToPreparedStatement( getDBReadyValueFromBean(bean), ps, psIndex);
	}

	/**
	 * Extract the value from a {@link ResultSet} and set it to a bean's property
	 * @param bean
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws SQLException
	 */
	public void getFromResultSet(final BEAN bean, final ResultSet rs, final String rsColumnName) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, SQLException {
		setDBValueToBean( bean, this.jdbcIO.getValueFromResultSet(rs, rsColumnName) );
	}

	/**
	 * Extract the value from a {@link ResultSet} and set it to a bean's property
	 * @param bean
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 * @throws SQLException
	 */
	public void getFromResultSet(final BEAN bean, final ResultSet rs, final int rsColumnIndex) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException, SQLException {
		setDBValueToBean( bean, this.jdbcIO.getValueFromResultSet(rs, rsColumnIndex) );
	}

	/**
	 * Set copy the property value from source to destination
	 * @return
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 * @throws IllegalArgumentException
	 */
	public void clonePropertyValue(final BEAN source, final BEAN destination) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		setPropertyValueToBean(destination, getPropertyValueFromBean(source));
	}

	public void increaseVersion(final BEAN bean, final boolean firstVersionNumber) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		setDBValueToBean(bean, this.math.increase(firstVersionNumber, getDBReadyValueFromBean(bean)));
	}

	public DB getDBReadyValueFromBean(final BEAN bean) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		return this.typeWrapper.unWrap(getPropertyValueFromBean(bean));
	}

	public void setDBValueToBean(final BEAN bean, final DB value) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		setPropertyValueToBean(bean, this.typeWrapper.wrap(value));
	}

	public P getPropertyValueFromBean(final BEAN bean) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		return this.getter.getValue(bean);
	}

	public void setPropertyValueToBean(final BEAN bean, final P value) throws IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		this.setter.setValue(bean, value);
	}
}
